from blocks.types.base import (Event,
                               EventOrEvents,
                               Processor,
                               AsyncProcessor,
                               AnyProcessor,
                               Source,
                               AsyncSource,
                               AnySource,
                               Block,
                               TypeOfProcessor)
from blocks.types.dto import ParallelEvent
